package org.softevo.ample;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URL;
import java.nio.channels.FileChannel;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.MissingResourceException;
import java.util.Stack;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.SuffixFileFilter;
import org.eclipse.core.internal.resources.Workspace;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchManager;
import org.eclipse.debug.core.IStreamListener;
import org.eclipse.debug.core.Launch;
import org.eclipse.debug.core.model.IProcess;
import org.eclipse.debug.core.model.IStreamMonitor;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.junit.ui.JUnitPlugin;
import org.eclipse.jdt.internal.junit.ui.TestRunInfo;
import org.eclipse.jdt.launching.IVMInstall;
import org.eclipse.jdt.launching.IVMRunner;
import org.eclipse.jdt.launching.JavaRuntime;
import org.eclipse.jdt.launching.SocketUtil;
import org.eclipse.jdt.launching.VMRunnerConfiguration;
import org.softevo.ample.eval.AnalysisException;
import org.softevo.ample.eval.AverageSequenceWeightEvaluator;
import org.softevo.ample.eval.XMLPatternFileReader;
import org.softevo.ample.instrument.AmpleInstrumenter;
import org.softevo.ample.instrument.InstrumentationConfiguration;
import org.softevo.ample.instrument.OptionsClassCreator;
import org.softevo.sibrelib.InstrumentationException;
import org.softevo.sibrelib.classfilter.ClassFilter;
import org.softevo.sibrelib.classfilter.DefaultFilter;
import org.softevo.sibrelib.source.ClassFile;
import org.softevo.sibrelib.source.ClassFileSource;
import org.softevo.sibrelib.source.DirectorySource;

/**
 * Control class for the instrumentation and launching the instrumented code.
 */
public class AmpleLauncher implements IStreamListener {
  
  /** The instrumentation configuration to be used. <BR>
   */
  private InstrumentationConfiguration config = null;
  
  /** The launchh configuration used to execute the virtual
   *  machine for running the tests. <BR>
   */
  private ILaunchConfiguration launchConfiguration = null;

  /**
   * Java project targeted by the JUnits tests and analysed by Ample.
   */
  private IJavaProject project;
  
  /**
   * Project-specific working area for the Ample plugin.
   * @see #project
   */
  private File pluginDir;
  
  /**
   * Project's output directories stored as full path names.
   */
  private ArrayList outputLocations;
  
  /**
   * External JAR files required by the project stored as full path names.
   * <p>The last two entries of this list are the names of the two JAR files that enable JUnit support in Eclipse.</p>
   */
  private ArrayList requiredJars;
  
  /** Implementation of <code>IStreamListener</code>. <BR>
   *  This method is called whenever the vm that runs the
   *  tests writes to standard or error output streams.
   */
  public void streamAppended(String text, IStreamMonitor monitor) {
    if (AmplePlugin.DEBUG) {
      System.out.println(text);
    }
   }
  
  /**
   * Initializes a new instance of <code>AmpleLauncher</code>.
   * 
   * @param project Targeted Java project.
   * @throws JavaModelException If an error occured while accessing resources of the targeted Java project.
   * @throws JUnitSupportException If locating Eclipse's JUnit support (the JAR files) has failed.
   * @throws NullPointerException ...
   */
  public AmpleLauncher(IJavaProject project, InstrumentationConfiguration config, 
                       ILaunchConfiguration launchConfiguration) throws JavaModelException, JUnitSupportException, NullPointerException {
    this.project = project;
    this.config = config;
    this.launchConfiguration = launchConfiguration;
    initLists();
    // TODO: Consider referenced projects
    
    String pluginIdentifier = AmplePlugin.getDefault().getBundle().getSymbolicName();
    IPath pluginPath = project.getProject().getWorkingLocation(pluginIdentifier);
    pluginDir = pluginPath.toFile();
  }
  
  /**
   * Clears the contents of Ample's working directory.
   */
  public void cleanWorkingDir() {
    if (! pluginDir.exists()) {
      pluginDir.mkdirs();
    } else if (pluginDir.isFile()) {
      pluginDir.delete();
      pluginDir.mkdirs();
    } else { // pluginDir.isDirectory()
      // Recursively delete all the files and subdirectories
      Stack subDirs = new Stack();
      subDirs.push(pluginDir);
      do {
        File currentDir = (File) subDirs.peek();
        boolean hasOnlyFiles = true;
        File[] files = currentDir.listFiles();
        for (int i = 0; i < files.length; ++i) {
          if (files[i].isFile()) {
            files[i].delete();
          } else { // files[i].isDirectory()
            hasOnlyFiles = false;
            subDirs.push(files[i]);
          }
        }
        if (hasOnlyFiles) {
          subDirs.pop();
          if (! subDirs.empty()) {
            currentDir.delete();
          }
        }
      } while (! subDirs.isEmpty());
    }
  }

  /** Calculates the names of all classes that shall be
   *  instrumented. <BR>
   *  @param source the class file source
   *  @return the set of all classes in the source
   */
  private HashSet getInstrumentClassNames(ClassFileSource source, ClassFilter filter) {
    Iterator classIterator;
    ClassFile classFile;
    HashSet result;
    
    result = new HashSet();
    classIterator = source.getClassIterator();
    while (classIterator.hasNext()) {
      classFile = (ClassFile) classIterator.next();
      if (filter.accept(classFile.getClassName())) {
        result.add(classFile.getClassName());
      }
    }
    return result;
  }
  
  private HashMap createClassNameToJavaTypeMap(HashSet instrumentClassNames) throws JavaModelException{
    HashMap result;
    String className, typeName, packageName;
    Iterator classNameIterator;
    IType type;
    
    result = new HashMap();
    classNameIterator = instrumentClassNames.iterator();
    while (classNameIterator.hasNext()) {
      className = classNameIterator.next().toString();
      if (className.indexOf(".") != -1) {
        packageName = className.substring(0, className.lastIndexOf("."));
        typeName = className.substring(className.lastIndexOf(".") + 1, className.length());
      } else {
        packageName = "";
        typeName = className;
      }
      if (AmplePlugin.DEBUG) {
        //System.out.println("Package name : " + packageName);
        //System.out.println("Type name: " + typeName);
      }
      try {
        type = project.findType(packageName, typeName);
      } catch (JavaModelException exception) {
        type = null;
      }
      
      result.put(className, type);
    }
    return result;
  }
  
  private void copyDirectoryRecursively(File sourceDirectory, File targetDirectory) throws IOException {
    File[] children;
    File targetFile;
    
    children = sourceDirectory.listFiles();
    for (int counter = 0; counter < children.length; counter++) {
      targetFile = new File(targetDirectory.getAbsolutePath() + File.separator + children[counter].getName());
      if (AmplePlugin.DEBUG) {
        //System.out.println("Copying " + children[counter] + " to " + targetFile);
      }
      if (children[counter].isFile()) {
        FileUtils.copyFile(children[counter], targetFile);
      } else {
        if (!targetFile.exists()) {
          targetFile.mkdirs();
        }
        copyDirectoryRecursively(children[counter], targetFile);
      }
    }
  }
  
  /**
   * Copies all the output files of the targeted project into the working directory.
   * 
   * @throws IOException If copying a file to Ample's working directory fails, e.g. due to name collision.
   */
  public void copyClassFiles() throws IOException {
    File sourceDirectory, targetDirectory;
    
    if (! pluginDir.exists()) {
      pluginDir.mkdirs();
    }
    targetDirectory = new File(pluginDir.getAbsolutePath());
    for (int i = 0; i < outputLocations.size(); ++i) {
      sourceDirectory = new File(outputLocations.get(i).toString());
      copyDirectoryRecursively(sourceDirectory, targetDirectory);
    }
  }
  
  /**
   * Instruments all the .class files in the working directory.
   * <p>Note that {@link #copyClassFiles()} should be called prior to calling this method.</p>
   * @return a hashmap of class names mapped to their java types
   */
  public HashMap instrument() throws InstrumentationException, IOException, ClassNotFoundException, CoreException {
    AmpleInstrumenter instrumenter;
    IVMInstall vmInstall;
    
    config.setIdentityMapFileName(pluginDir.getAbsolutePath() + File.separator + "identity.map");
    config.setOutputFileName(pluginDir.getAbsolutePath() + File.separator + "patterns.xml");
    vmInstall = JavaRuntime.computeVMInstall(launchConfiguration);
    if (vmInstall == null) {
      vmInstall = JavaRuntime.getDefaultVMInstall();
    }
    instrumenter = new AmpleInstrumenter(config, pluginDir, pluginDir, new DefaultFilter(), vmInstall);
    instrumenter.instrument();
    return createClassNameToJavaTypeMap(getInstrumentClassNames(new DirectorySource(pluginDir, pluginDir), new DefaultFilter()));
  }
  
  /** This method creates the options class and fills it with
   *  the values needed by the instrumentation configuration. <BR>
   *  @throws IOException if an exception occurs writing the class
   */
  protected void writeOptionsClass() throws IOException {
    OptionsClassCreator classCreator;
    byte[] optionsClassBytes;
    File optionsClass;
    FileOutputStream fileOutputStream = null;
    
    classCreator = new OptionsClassCreator(config);
    optionsClassBytes = classCreator.create();
    optionsClass = new File(pluginDir.getAbsolutePath() + File.separator + "org" + 
                            File.separator + "softevo" + File.separator + "ample" + File.separator + 
                            "sibrelib" + File.separator + "runtime" + File.separator + "Options.class");
    try {
      fileOutputStream = new FileOutputStream(optionsClass);
      fileOutputStream.write(optionsClassBytes);
      fileOutputStream.flush();
    } finally {
      if (fileOutputStream != null) {
        try {
          fileOutputStream.close();
        } catch (IOException finalizeException) {
          //ignore exception
        }
      }
    }
  }
  
  /** This method is currently not used. <BR>
   * @param sourceDirectory
   * @throws IOException
   */
  protected void copyFromJar(String sourceDirectory) throws IOException {
    File jarFile;
    JarFile ampleJar;
    Enumeration entries;
    ZipEntry entry;
    
    
    jarFile = new File(sourceDirectory + File.separator + "ample.jar");
    if (jarFile.exists()) {
      ampleJar = new JarFile(jarFile);
      entries = ampleJar.entries();
      while (entries.hasMoreElements()) {
        entry = (ZipEntry) entries.nextElement();
        System.out.println(entry.getName());
        if (entry.getName().startsWith("org/softevo/ample/sibrelib/runtime")) {
          
        }
      }
    } else {
      throw new FileNotFoundException(sourceDirectory + File.separator + "ample.jar");
    }
  }
  
  /** This method copies all runtime classes needed for the execution of AMPLE. <BR>
   *  This includes all sibrelib jars and required apache files. The method also adds
   *  entries for those jar files in the class path. 
   *  @param sourceDirectory the directory where the classes are located (the plugin directory of ample)
   *  @throws IOException if an exception occurs copying classes
   */
  protected void copyAmpleRuntimeClasses(String sourceDirectory) throws IOException {
    File targetDirectory, sourceRuntimeClassDirectory, sibreLibJar, commonsLangJar, runtimeDir;
    File[] children;
    String binDir;
    
    //if we are running a deployed version, we need to copy classes from runtime directory
    //in our plugin folder
    runtimeDir = new File(sourceDirectory + File.separator + "runtime");
    if (runtimeDir.exists()) {
      binDir = "runtime";
    } else {
      binDir = "bin";
    }
    if (AmplePlugin.DEBUG) {
      System.out.println(pluginDir);
    }
    targetDirectory = new File(pluginDir.getAbsolutePath() + File.separator + "org" + 
                               File.separator + "softevo" + File.separator + "ample" + File.separator + 
                               "sibrelib" + File.separator + "runtime");
    targetDirectory.mkdirs();
    sourceRuntimeClassDirectory = new File(sourceDirectory + File.separator + binDir + File.separator + "org" + 
                                           File.separator + "softevo" + File.separator + "ample" + File.separator + 
                                           "sibrelib" + File.separator + "runtime");
    children = sourceRuntimeClassDirectory.listFiles((FilenameFilter) new SuffixFileFilter("class"));
    for (int counter = 0; counter < children.length; counter++) {
      FileUtils.copyFileToDirectory(children[counter], targetDirectory);
    }
    sibreLibJar = new File(sourceDirectory + File.separator + "lib" + File.separator + "sibrelib.jar");
    commonsLangJar = new File(sourceDirectory + File.separator + "lib" + File.separator + "commons-lang-2.0.jar");
    requiredJars.add(sibreLibJar.toString());
    requiredJars.add(commonsLangJar.toString());
    writeOptionsClass();
  }
  
  protected String findJUnitJarInPlugins() throws MalformedURLException, IOException {
    IPath junitJar;
    File jarFile;
    IClasspathEntry entry;
    
    IPath junitHome= new Path(JUnitPlugin.JUNIT_HOME);
    IPath sourceHome= new Path("ECLIPSE_HOME"); //$NON-NLS-1$
    entry= JavaCore.newVariableEntry(
        junitHome.append("junit.jar"),   //$NON-NLS-1$
        //TODO: find a better solution than declaring a classpath variable
        sourceHome.append("plugins/org.eclipse.jdt.source_3.1.0/src/org.junit_3.8.1/junitsrc.zip"),  //$NON-NLS-1$
        null
    );
    if (AmplePlugin.DEBUG) {
      System.out.println(entry);
      System.out.println(entry.getPath().toOSString());
      System.out.println(entry.getResolvedEntry().getPath().toOSString());
    }
    return entry.getResolvedEntry().getPath().toOSString();
  }
  
  /**
   * This method prepares and copies all runtime classes needed by ample.
   * <p>It should be called after {@link #instrument()}.</p>
   */
  public void prepareRuntimeClasses() throws IOException, MalformedURLException {
    URL url;
    String rootDirectory;
    
    url = Platform.getBundle("org.softevo.ample").getEntry("/");;
    rootDirectory = Platform.asLocalURL(url).getFile();
    if (AmplePlugin.DEBUG) {
      System.out.println(rootDirectory);
    }
    copyAmpleRuntimeClasses(rootDirectory);
  }
  
  /** This method executes a junit test. <BR>
   *  @param testRun the test to be executed
   *  @param vm the virtual machine to run the test
   *  @return the reader for the pattern file that was recorded or <code>null</code> if something went wrong
   *  @throws CoreException if an exception occurs launching the process
   *  @throws AnalysisException if an exception occurs reading the pattern file
   *  @throws MissingResourceException if a workspace ressource cannot be found
   */
  protected XMLPatternFileReader runTest(TestRunInfo testRun,  IVMInstall vm) throws CoreException, AnalysisException, MissingResourceException {
    int port;
    int maxNumberOfWaits = AmplePlugin.getDefault().getMaxTestDurationSeconds() * 2;
    IProcess[] processes;
    File lockFile;
    XMLPatternFileReader patternFileReader;
    
    lockFile = new File(pluginDir.getAbsolutePath() + File.separator + 
                        ".lock");
    if (lockFile.exists()) {
      lockFile.delete();
    }
    // Get IVMRunner
    IVMRunner runner = vm.getVMRunner(ILaunchManager.RUN_MODE);
    
    // Create VMRunnerConfiguration
    port = SocketUtil.findFreePort();
    if (port == -1) {
      throw new MissingResourceException(null, "SocketUtil", null);
    }
    String testName = testRun.getClassName() + ":" + testRun.getTestMethodName();
    VMRunnerConfiguration config = createConfig(testName, port);
    
    // Run IVMRunner with the configuration
    ILaunch launch = new Launch(null, ILaunchManager.RUN_MODE, null);
    if (AmplePlugin.DEBUG) {
      System.out.print("Running " + testName + "... ");
    }
    TestListenerFaker faker = new TestListenerFaker(port);
    new Thread(faker).start();
    runner.run(config, launch, null);
    try {
      Thread.sleep(100);
    } catch (InterruptedException wakeUp) {
      //ignore exception
    }
    launch.getProcesses()[0].getStreamsProxy().getErrorStreamMonitor().addListener(this);
    launch.getProcesses()[0].getStreamsProxy().getErrorStreamMonitor().addListener(this);
    for (int counter = 0; counter < maxNumberOfWaits; counter++) {
      try {
        Thread.sleep(500);
      } catch (InterruptedException wakeUp) {
        //ignore exception
      }
      if (lockFile.exists()) {
        break;
      }
    }
    if (lockFile.exists()) {
      patternFileReader = new XMLPatternFileReader(pluginDir.getAbsolutePath() + File.separator + "patterns.xml");
      patternFileReader.parseFile();
      return patternFileReader;
    } else {
      launch.getProcesses()[0].terminate();
      throw new AnalysisException(MessageFormat.format(AmpleMessages.ErrorTimeout_message, new String[] {testName}));
    }
  }
  
  /**
   * Reruns the JUnit tests requested by the user on the instrumented byte code.
   * <p>Note that {@link #instrument()} should be called prior to calling this method.</p>
   * 
   * @param junitLaunchConfig Launch configuration of the JUnit tests run before running Ample.
   * @param testRuns Array of test runs to be launched.
   * @throws CoreException ...
   */
  public AverageSequenceWeightEvaluator launch(TestRunInfo failingRun, TestRunInfo[] passingRuns, IProgressMonitor monitor) throws CoreException, MissingResourceException, AnalysisException {
    XMLPatternFileReader failingPatterns, passing;
    Vector passingPatterns;
    
    passingPatterns = new Vector(passingRuns.length);
    // Get VMInstall
    IVMInstall vm = JavaRuntime.computeVMInstall(launchConfiguration);
    if (vm == null) {
      vm = JavaRuntime.getDefaultVMInstall();
    }
    monitor.subTask(AmpleMessages.AmpleLauncher_subtask_execfailing + " " + failingRun.getTestName());
    failingPatterns = runTest(failingRun, vm);
    if (failingPatterns == null) {
      throw new AnalysisException("Unable to trace execution of failing run for test " + failingRun.getTestName() + ".");
    }
    monitor.worked(1);
    for (int i = 0; i < passingRuns.length; ++i) {
      monitor.subTask(AmpleMessages.AmpleLauncher_subtask_execpassing + " " + passingRuns[i].getTestName());
      passing = runTest(passingRuns[i], vm);
      if (passing == null) {
        throw new AnalysisException("Unable to trace execution of failing run for test " + passingRuns[i].getTestName() + ".");
      }
      passingPatterns.add(passing);
      monitor.worked(1);
    }
    return new AverageSequenceWeightEvaluator(failingPatterns, passingPatterns);
  }
  
  /** A fake class that is used to convince the remote JUnit-Runner of
   *  Eclipse that everything is okay. <BR>
   *  This class simply opens a socket on a predefined port and reads data
   *  until the stream is closed by the remote side. No JUnit test will run
   *  without this.
   *  @author dallmeier
   */
  private class TestListenerFaker implements Runnable {
    
    /** The port we are listening on. <BR>
     */
    private int port = -1;
    
    /** Creates a new listener. <BR>
     *  @param port the port we listen on
     */
    public TestListenerFaker(int port) {
      this.port = port;
    }
    
    /** Opens a <code>ServerSocket</code>, waits for
     *  a single incoming connection and reads data from
     *  this connection forever. <BR>
     */
    public void run() {
      ServerSocket serverSocket = null;
      Socket connection = null;
      InputStream inputStream;
      BufferedReader reader;
      String message;
      
      try {
        serverSocket = new ServerSocket(port);
        connection = serverSocket.accept();
        if (AmplePlugin.DEBUG) {
          System.out.println("Accepted connection");
        }
        try {
          inputStream = connection.getInputStream();
          reader = new BufferedReader(new InputStreamReader(inputStream));
          do {
            message = reader.readLine();
            if (AmplePlugin.DEBUG) {
              System.out.println("Received message " + message);
            }
          } while (message != null);
        } catch (IOException readException) {
          //this indicates that the remote test runner has finished
          if (AmplePlugin.DEBUG) {
            System.out.println("Exception reading from remote test runner.");
            readException.printStackTrace();
          }
        }
      } catch (Exception listenException) {
        if (AmplePlugin.DEBUG) {
          System.out.println("Exception listening for test runner messages.");
          listenException.printStackTrace();
        }
      } finally {
        if (serverSocket != null) {
          try {
            serverSocket.close();
          } catch (IOException finalizeException) {
            //ignore exception
          }
        }
        if (connection != null) {
          try {
            connection.close();
          } catch (IOException finalizeException) {
            //ignore exception
          }
        }
      }
    }
  }

  /** A <code>ProgressMonitor</code> that simply logs to <code>System.out</code>
   */
  private class SysOutMonitor implements IProgressMonitor {
    
    /** Flag if the monitored task is canceled. <BR>
     */
    private boolean isCanceled = false;
    
    /** Callback if a task has begun. <BR>
     */
    public void beginTask(String name, int totalWork) {
      System.out.println("Beginning task " + name);
    }

    /** Callback if the task is finished. <BR>
     */
    public void done() {
      System.out.println("Done.");
    }
    
    /** Not implemented. <BR>
     */
    public void internalWorked(double work) {
      // TODO Auto-generated method stub
    }
    
    public boolean isCanceled() {
      return isCanceled;
    }
    
    public void setCanceled(boolean value) {
      isCanceled = value;
    }
    
    public void setTaskName(String name) {
      System.out.println("Setting task name " + name);
    }
    
    public void subTask(String name) {
      System.out.println("Subtask " + name);
    }
    
    public void worked(int work) {
      System.out.println("Worked " + work);
    }
    
  }
  /**
   * Copies a specified file to a given location.
   * 
   * @param file File to be copied. If the specified file does not exist, calling this method has no effect.
   * @param targetDir Location of the file.
   * @throws IOException When I/O error occurs.
   */
  private static void copyFile(File file, File targetDir) throws IOException {
    try {
      String targetFileName = targetDir.getAbsolutePath() + File.separator + file.getName();
      FileChannel srcChannel = new FileInputStream(file).getChannel();
      FileChannel destChannel = new FileOutputStream(targetFileName).getChannel();
      destChannel.transferFrom(srcChannel, 0, srcChannel.size());
      srcChannel.close();
      destChannel.close();
    } catch (FileNotFoundException e) {
      // Ignore
    }
  }
  
  /**
   * Adds to the given list the full names of the two JAR files that enable JUnit support in Eclipse.
   * <p>Note: This method contains code copied from <code>JUnitLaunchConfiguration.createClassPath(ILaunchConfiguration)</code>.</p>
   * 
   * @param jarsList List of JAR file names.
   * @throws JUnitSupportException If locating any of the two JAR files has failed.
   * @throws NullPointerException If <code>jarsList</code> is <code>null</code>.
   */
  private void addJUnitJars(ArrayList jarsList) throws JUnitSupportException {
    final String bin = "bin"; //$NON-NLS-1$
    final String supportJar = "junitsupport.jar"; //$NON-NLS-1$
    final String runtimeJar = "junitruntime.jar"; //$NON-NLS-1$
    
    URL url = Platform.getBundle(JUnitPlugin.PLUGIN_ID).getEntry("/"); //$NON-NLS-1$
    URL runtimeURL = Platform.getBundle("org.eclipse.jdt.junit.runtime").getEntry("/"); //$NON-NLS-1$ //$NON-NLS-2$
    
    String support = null; // full name of the JAR file for JUnit support
    String runtime = null; // full name of the JAR file for JUnit runtime in Eclipse
    String jUnit = null;
    
    try {
      if (Platform.inDevelopmentMode()) {
        // Platform in development mode, first try the bin output folder
        try {
          support = Platform.asLocalURL(new URL(url, bin)).getFile();
        } catch (IOException e1) {
          support = Platform.asLocalURL(new URL(url, supportJar)).getFile();
        }
        try {
          runtime = Platform.asLocalURL(new URL(runtimeURL, bin)).getFile();
        } catch (IOException e2) {
          runtime = Platform.asLocalURL(new URL(runtimeURL, runtimeJar)).getFile();
        }
        
      } else { // ! Platform.inDevelopmentMode()
        support = Platform.asLocalURL(new URL(url, supportJar)).getFile();
        runtime = Platform.asLocalURL(new URL(runtimeURL, runtimeJar)).getFile();
      }
      
      jarsList.add(support);
      jarsList.add(runtime);
      jUnit = findJUnitJarInPlugins();
      if (jUnit != null) {
        jarsList.add(jUnit);
      }
      if (AmplePlugin.DEBUG) {
        System.out.println("Using JUnit jar \"" + jUnit + "\"");
      }
    } catch (IOException e) {
      throw new JUnitSupportException();
    }
  }     
  
  /**
   * Initializes the two fields - {@link #outputLocations} and {@link #requiredJars}.
   * 
   * @throws JavaModelException If an error occured while accessing resources of the targeted Java project.
   * @throws JUnitSupportException If locating Eclipse's JUnit support (the JAR files) has failed.
   * @throws NullPointerException ...
   */
  private void initLists() throws JavaModelException, JUnitSupportException, NullPointerException {
    IFile jarFile;
    IPath jarFilePath;
    
    outputLocations = new ArrayList();
    requiredJars = new ArrayList();
    
    IPath outputPathRel = project.readOutputLocation().makeRelative();
    String firstSegment = outputPathRel.segment(0);
    if (firstSegment != null && firstSegment.equals(project.getElementName())) {
      outputPathRel = outputPathRel.removeFirstSegments(1);
    }
    IFolder outputFolder = project.getProject().getFolder(outputPathRel);
    outputLocations.add(outputFolder.getLocation().toOSString());
    IClasspathEntry[] classPaths = project.getResolvedClasspath(true);
    for (int i = 0; i < classPaths.length; ++i) {
      IClasspathEntry cPath = classPaths[i];
      if (cPath.getContentKind() == IPackageFragmentRoot.K_BINARY) {
        if (cPath.getEntryKind() == IClasspathEntry.CPE_LIBRARY) {
          // Add to the list of required JARs
          jarFilePath = cPath.getPath().makeRelative();
          firstSegment = jarFilePath.segment(0);
          if (firstSegment != null && firstSegment.equals(project.getElementName())) {
            jarFilePath = jarFilePath.removeFirstSegments(1);
          }
          jarFile = project.getProject().getFile(jarFilePath);
          if (AmplePlugin.DEBUG) {
            System.out.println("Adding dependant jar " + jarFile.getLocation().toOSString() + " to classpath.");
          }
          requiredJars.add(jarFile.getLocation().toOSString());
        }
      } else { // cPath.getContentKind() == IPackageFragmentRoot.K_SOURCE
        IPath outputLoc = cPath.getOutputLocation();
        if (outputLoc != null) {
          // TODO: Add the output location using IProject.getFolder(...)?
          outputLocations.add(outputLoc.toOSString());
        }
      }
    }
    
    addJUnitJars(requiredJars);
  }
  
  /**
   * Computes the classpath using the directory with the instrumented .class files and the required JAR files.
   * 
   * @return Classpath in the form of <code>String</code> array.
   */
  private String[] createClassPath() {
    final int requiredJarsCount = requiredJars.size();
    String[] classPath = new String[1 + requiredJarsCount];
    classPath[0] = pluginDir.getAbsolutePath();
    for (int i = 0; i < requiredJarsCount; ++i) {
      classPath[i + 1] = (String) requiredJars.get(i);
    }
    return classPath;
  }
  
  /**
   * Creates a configuration with class path that includes <code>org.eclipse.jdt.junit</code> plugin.
   * In addition it adds the port for the RemoteTestRunner as an argument.
   * <p>Note: This method contains code copied from
   * <code>JUnitLaunchConfiguration.createVMRunner(ILaunchConfiguration, IType[], int)</code>.</p>
   * 
   * @param testName Name of test to be run. It has the form {classname}<code>:<code>{methodname}.
   * @throws CoreException ...
   * @throws MissingResourceException If no free port could be found.
   */
  protected VMRunnerConfiguration createConfig(String testName, int port) throws CoreException, MissingResourceException {
    final String mainType = "org.eclipse.jdt.internal.junit.runner.RemoteTestRunner"; //$NON-NLS-1$
    String[] classPath = createClassPath();
    if (AmplePlugin.DEBUG) {
      System.out.print("Classpath: ");
      for (int i = 0; i < classPath.length; ++i) {
        System.out.print("\t" + classPath[i]);
      }
      System.out.println();
      System.out.println("Working directory: " + pluginDir.getAbsolutePath());
    }
    VMRunnerConfiguration vmConfig = new VMRunnerConfiguration(mainType, classPath);
    
    // Insert the program arguments
    String[] args = new String[6];
    args[0] = "-version"; //$NON-NLS-1$
    args[1] = "3"; //$NON-NLS-1$
    args[2] = "-port"; //$NON-NLS-1$
    args[3] = String.valueOf(port);
    args[4] = "-test"; //$NON-NLS-1$
    args[5] = testName; //$NON-NLS-1$
    
    vmConfig.setWorkingDirectory(pluginDir.getAbsolutePath());
    vmConfig.setProgramArguments(args);
    return vmConfig;
  }
  
}
