package org.softevo.ample;

import java.io.IOException;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.MissingResourceException;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.junit.ui.TestRunInfo;
import org.eclipse.jdt.internal.junit.ui.TestRunnerViewPart;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IViewActionDelegate;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.softevo.ample.eval.AnalysisException;
import org.softevo.ample.eval.AverageSequenceWeightEvaluator;
import org.softevo.ample.eval.ClassNameWithWeight;
import org.softevo.ample.instrument.InstrumentationConfiguration;
import org.softevo.ample.ui.SelectTestsDialog;

/**
 * Behavior of Ample's item in the context menu of the Failure tab in the JUnit Results view.
 */
public class PopupMenuAction implements IViewActionDelegate {

  /**
   * Instance of the JUnit Results view part.
   */
  private TestRunnerViewPart viewPart;
  
  /**
   * Test currently selected in the active view tab.
   */
  private TestRunInfo selectedTest;
  
  /* (non-Javadoc)
   * @see org.eclipse.ui.IViewActionDelegate#init(org.eclipse.ui.IViewPart)
   */
  public void init(IViewPart view) {
    viewPart = (TestRunnerViewPart) view;
  }
  
  /* (non-Javadoc)
   * @see org.eclipse.ui.IActionDelegate#run(org.eclipse.jface.action.IAction)
   */
  public void run(IAction action) {
    Shell shell = viewPart.getSite().getShell();
    
    // Verify there are no JUnit tests currently running
    if (AmplePlugin.getSharedListener().areTestsRunning()) {
      showErrorTestsAreRunning(shell);
      return;
    }
    
    // Verify a failing test is selected
    int index = AmplePlugin.getSharedListener().indexOfFailing(selectedTest);
    if (index == -1) {
      // Ample clicked when not a failing test was selected
      showErrorFailingRequired(shell);
      return;
    }
    
    // Verify enough passing tests are present
    int passingTestsNum = AmplePlugin.getSharedListener().getPassingTestsCount();
    if (passingTestsNum < AmpleSettings.MIN_PASSING_TESTS) {
      showErrorTooFewPassing(shell);
      return;
    }
    
    // Display dialog to choose which tests to be analyzed
    SelectTestsDialog d = new SelectTestsDialog(shell, index);
    d.setBlockOnOpen(true);
    d.open();
    
    if (d.getReturnCode() == Window.OK) {
      // Read client's choice on instrumentation of failing and passing tests
      TestRunInfo[] passingTests = d.getSelectedPassingTests();
      TestRunInfo[] testRuns = new TestRunInfo[1 + passingTests.length];
      testRuns[0] = d.getSelectedFailingTest();
      for (int i = 0; i < passingTests.length; ++i) {
        testRuns[i + 1] = passingTests[i];
      }
      // Launch Ample with the selected failing and passing test runs
      launchAmple(d.getSelectedFailingTest(), d.getSelectedPassingTests());
    }
  }
  
  /* (non-Javadoc)
   * @see org.eclipse.ui.IActionDelegate#selectionChanged(org.eclipse.jface.action.IAction,
   *      org.eclipse.jface.viewers.ISelection)
   */
  public void selectionChanged(IAction action, ISelection selection) {
    // Retrieve the selected element
    if (selection instanceof IStructuredSelection) {
      // Even when StructuredSelection is passed, there is a single element selected
      selectedTest = (TestRunInfo) ((IStructuredSelection) selection).getFirstElement();
    } else {
      selectedTest = (TestRunInfo) selection;
    }
//  System.out.println(selectedTest.getTestId() + " " + selectedTest.getTestName());
  }
  
  private void blankRankingView() {
    if (AmplePlugin.getRankingView() != null) {
      Display.getDefault().syncExec(new Runnable() { 
        public void run ()  {
          AmplePlugin.getRankingView().showBlank();
        }
      });
    }
  }

  /**
   * Instruments project's byte code files and reruns the tests.
   * 
   * @param testRuns Array of test runs to be launched.
   */
  private void launchAmple(final TestRunInfo failingRun, final TestRunInfo[] passingRuns) {
    Job runAmpleJob;
    
    try {
      PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().showView("org.softevo.ample.ui.views.RankingView");
    } catch (Exception viewException) {
      viewException.printStackTrace();
    }
    runAmpleJob = new Job("AMPLE Test Execution") {
      protected IStatus run(IProgressMonitor monitor) {
        final AverageSequenceWeightEvaluator evaluator;
        final ClassNameWithWeight[] ranking;
        final HashMap classNameToJavaTypeMap;
        IJavaProject project;
        final InstrumentationConfiguration config;
        
        project = viewPart.getLaunchedProject();
        // TODO: Verify, if needed, that all required projects were built
        // TODO: Make all dialogs for the errors and put them in the resource bundle
        try {
          monitor.beginTask(AmpleMessages.PopupMenuAction_task, 6 + passingRuns.length);
          if (AmplePlugin.getRankingView() != null) {
            Display.getDefault().syncExec(new Runnable() { 
              public void run ()  {
                AmplePlugin.getRankingView().showRunning();
              }
            });
          }
          config = AmplePlugin.getDefault().getInstrumentationConfiguration();
          AmpleLauncher launcher = new AmpleLauncher(project, config, viewPart.getLastLaunch().getLaunchConfiguration());
          monitor.subTask(AmpleMessages.PopupMenuAction_subtask_clean);
          launcher.cleanWorkingDir();
          monitor.worked(1);
          monitor.subTask(AmpleMessages.PopupMenuAction_subtask_copy);
          launcher.copyClassFiles();
          monitor.worked(1);
          monitor.subTask(AmpleMessages.PopupMenuAction_subtask_instrument);
          classNameToJavaTypeMap = launcher.instrument();
          monitor.worked(1);
          launcher.prepareRuntimeClasses();
          monitor.worked(1);
          monitor.subTask(AmpleMessages.PopupMenuAction_subtask_exec);
          evaluator = launcher.launch(failingRun, passingRuns, monitor);
          monitor.subTask(AmpleMessages.PopupMenuAction_subtask_calculate);
          monitor.worked(1);
          ranking = evaluator.calculateRanking();
          if (AmplePlugin.getRankingView() != null) {
            Display.getDefault().syncExec(new Runnable() { 
              public void run ()  {
                AmplePlugin.getRankingView().setInput(ranking, classNameToJavaTypeMap, evaluator.getClassNameToPatternDifferencesMap(), config);
              }
            });
          }
          return new Status(IStatus.OK, "org.softevo.ample", IStatus.OK, "Done.", null);
        } catch (final AnalysisException exception) {
          System.out.println("Analysis exception: " + exception);
          blankRankingView();
          Display.getDefault().syncExec(new Runnable() { 
            public void run ()  {
              MessageDialog.openInformation(viewPart.getSite().getShell(), AmpleMessages.ErrorTracing_dialog_title,
                                            exception.getMessage());
            }
          });
          return new Status(IStatus.OK, "org.softevo.ample", IStatus.OK, "Done.", null);
          //return new Status(IStatus.ERROR, "org.softevo.ample", IStatus.OK, AmpleMessages.PopupMenuAction_error_launch, exception);
        } catch (Exception e) {
          blankRankingView();
          return new Status(IStatus.ERROR, "org.softevo.ample", IStatus.OK, AmpleMessages.PopupMenuAction_error_launch, e);
        } finally {
          monitor.done();
        }
      }
    };
    runAmpleJob.setPriority(Job.LONG);
    runAmpleJob.schedule();
  }
  
  /**
   * Displays an error message dialog stating Ample should be run when all JUnits have ended.
   * 
   * @param shell Owner of the message dialog that will appear.
   */
  private static void showErrorTestsAreRunning(Shell shell) {
    MessageDialog.openError(shell, AmpleMessages.ErrorTestsAreRunning_dialog_title,
                            AmpleMessages.ErrorTestsAreRunning_message);
  }
  
  /**
   * Displays an error message dialog stating Ample should be run when a failing test is selected.
   * 
   * @param shell Owner of the message dialog that will appear.
   */
  private static void showErrorFailingRequired(Shell shell) {
    MessageDialog.openError(shell, AmpleMessages.ErrorFailingRequired_dialog_title,
                            AmpleMessages.ErrorFailingRequired_message);
  }
  
  /**
   * Displays an error message dialog stating there are too few passing passing tests for Ample to
   * run.
   * 
   * @param shell Owner of the message dialog that will appear.
   */
  private static void showErrorTooFewPassing(Shell shell) {
    String message;
    if (AmpleSettings.MIN_PASSING_TESTS > 1) {
      String[] msgParams = new String[] { String.valueOf(AmpleSettings.MIN_PASSING_TESTS) };
      message = MessageFormat.format(AmpleMessages.ErrorTooFewPassing_message, msgParams);
    } else {
      message = AmpleMessages.ErrorTooFewPassing_message1;
    }
    MessageDialog.openError(shell, AmpleMessages.ErrorTooFewPassing_dialog_title, message);
  }
  
}
