package org.softevo.ample.instrument;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;
import java.util.zip.ZipFile;

import org.eclipse.core.runtime.IPath;
import org.eclipse.jdt.launching.IVMInstall;
import org.eclipse.jdt.launching.JavaRuntime;
import org.eclipse.jdt.launching.LibraryLocation;
import org.softevo.ample.AmplePlugin;
import org.softevo.ample.sibrelib.extensions.IncomingCallTracer;
import org.softevo.ample.sibrelib.extensions.OutgoingCallTracer;
import org.softevo.ample.sibrelib.runtime.IncomingClassSequenceWriter;
import org.softevo.sibrelib.InstrumentationContext;
import org.softevo.sibrelib.InstrumentationException;
import org.softevo.sibrelib.TraceInstrumenter;
import org.softevo.sibrelib.classfilter.ClassFilter;
import org.softevo.sibrelib.source.ClassFile;
import org.softevo.sibrelib.source.ClassFileSource;
import org.softevo.sibrelib.source.DirectorySource;
import org.softevo.sibrelib.source.JarSource;
import org.softevo.sibrelib.util.IdentifierMap;
import org.softevo.sibrelib.util.IdentifierMapGenerator;


/** This class contains contains the entry point for
 *  the instrumentation of bytecode.
 */
public class AmpleInstrumenter  {
  
  /** Instrumentation configuration for instrumenter run. <BR>
   */
  private InstrumentationConfiguration config = null;
  
  /** The directory used as working directory by the virtual machine instance. <BR>
   */
  private File workingDirectory = null;
  
  /** The directory that contains the classes that should be instrumented. <BR>
   */
  private File classDirectory = null;
  
  /** The filter to be used. <BR>
   */
  private ClassFilter filter = null;
  
  /** The virtual machine used for running instrumented files. <BR>
   */
  private IVMInstall vmInstall = null;
  
  /** The set of class names that are instrumented. <BR>
   */
  private HashSet instrumentClassNames = null;
  
  /** Creates a new instrumenter for a set of classes. <BR>
   *  @param config the instrumentation configuration
   *  @param classDirectory the class directory
   *  @param workingDirectory the working directory of the virtual machine 
   *  @param filter a filter for classes that shouldn't be instrumented
   *  @param vmInstall the virtual machine
   */
  public AmpleInstrumenter(InstrumentationConfiguration config, File classDirectory,
                           File workingDirectory, 
                           ClassFilter filter,
                           IVMInstall vmInstall) {
    this.config = config;
    this.classDirectory = classDirectory;
    this.workingDirectory = workingDirectory;
    this.filter = filter;
    this.vmInstall = vmInstall;
  }
  
  /** Determines which instrumentation extensions are needed by
   *  the given configuration. <BR>
   *  @param context the instrumentation context
   *  @return the list of trace extensions
   */
  private Vector getTraceExtensions(InstrumentationContext context) {
    Vector result;
    
    result = new Vector();
    if (config.isClassLevel()) {
      if (config.isIncomingCalls()) {
        result.add(new IncomingCallTracer(context, "org.softevo.ample.sibrelib.runtime.IncomingClassSequenceWriter", false));
      } else {
        result.add(new OutgoingCallTracer(context, "org.softevo.ample.sibrelib.runtime.OutgoingClassSequenceWriter", false));
      }
    } else {
      if (config.isIncomingCalls()) {
        result.add(new IncomingCallTracer(context, "org.softevo.ample.sibrelib.runtime.IncomingObjectSequenceWriter", true));
      } else {
        result.add(new OutgoingCallTracer(context, "org.softevo.ample.sibrelib.runtime.OutgoingObjectSequenceWriter", true));
      }
    }
    return result;
  }
  
  /** Calculates the names of all classes that shall be
   *  instrumented. <BR>
   *  @param source the class file source
   *  @return the set of all classes in the source
   */
  private HashSet getInstrumentClassNames(ClassFileSource source) {
    Iterator classIterator;
    ClassFile classFile;
    HashSet result;
    
    result = new HashSet();
    classIterator = source.getClassIterator();
    while (classIterator.hasNext()) {
      classFile = (ClassFile) classIterator.next();
      if (filter.accept(classFile.getClassName())) {
        result.add(classFile.getClassName());
      }
    }
    return result;
  }

  /** Saves a set of core classes into the cache directory. <BR>
   *  @param cacheFile the file the cached core class set should be stored in
   *  @param coreClassSet the class set to cache
   *  @throws IOException if an exception occurs writing the cache file
   */
  private void writeToCache(File cacheFile, HashSet coreClassSet) throws IOException {
    ObjectOutputStream objectOutputStream;
    FileOutputStream fileOutputStream = null;
    
    try {
      fileOutputStream = new FileOutputStream(cacheFile);
      objectOutputStream = new ObjectOutputStream(fileOutputStream);
      objectOutputStream.writeObject(coreClassSet);
    } finally {
      if (fileOutputStream != null) {
        try {
          fileOutputStream.close();
        } catch (IOException finalizeException) {
          //ignore exception
        }
      }
    }
  }
  
  /** This method tries to read the core class set from the cache. <BR>
   *  @param cacheFile the file that is supposed to contain the cached values
   *  @return the core class set or <code>null</code> if not found
   *  @throws IOException if an exception occurs reading from the cache
   */
  private HashSet readFromCache(File cacheFile) throws IOException, ClassNotFoundException {
    ObjectInputStream objectInputStream;
    FileInputStream fileInputStream = null;

    try {
      fileInputStream = new FileInputStream(cacheFile);
      objectInputStream = new ObjectInputStream(fileInputStream);
      return (HashSet) objectInputStream.readObject();
    } finally {
      if (fileInputStream != null) {
        try {
          fileInputStream.close();
        } catch (IOException finalizeException) {
          //ignore exception
        }
      }
    }
  }
  
  /** Gets the set of core classes to be used by te instrumentation. <BR>
   *  This method tries to locate the runtime class set from the virtual machine installation.
   *  If this succeeds, the method tries to locate the set in the cache. If this fails, the
   *  runtime jar file is analysed in order to find all the classes. 
   *  @return the core class set
   *  @throws InstrumentationException if the runtime jar file was not found 
   *  @throws IOException if an exception occurs reading from or writing to the cache
   */
  private HashSet getCoreClassSet() throws InstrumentationException, IOException, ClassNotFoundException {
    LibraryLocation[] libraryLocations;
    IPath jreJarPath = null, cacheDirectory = null;
    HashSet result;
    Iterator classIterator;
    JarSource source;
    ClassFile classFile;
    File cachedFile;
    
    libraryLocations = vmInstall.getVMInstallType().getDefaultLibraryLocations(vmInstall.getInstallLocation());
    for (int counter = 0; counter < libraryLocations.length; counter++) {
      if (libraryLocations[counter].getSystemLibraryPath() != null) {
        jreJarPath = libraryLocations[counter].getSystemLibraryPath();
        break;
      }
    }
    if (jreJarPath == null) {
      throw new InstrumentationException("Unable to locate runtime jar.");
    } else {
      cacheDirectory = AmplePlugin.getDefault().getStateLocation();
      cachedFile = new File(cacheDirectory.toOSString() + File.separator + 
                            jreJarPath.toString().replace(File.separatorChar, '.').replace(':', '.'));
      System.out.println("Checking cache in " + cachedFile.getAbsolutePath());
      if (cachedFile.exists()) {
        return readFromCache(cachedFile);
      } else {
        result = new HashSet(1000);
        source = new JarSource(new ZipFile(jreJarPath.toFile()),
                               new File(""));
        classIterator = source.getClassIterator();
        while (classIterator.hasNext()) {
          classFile = (ClassFile) classIterator.next();
          result.add(classFile.getClassName());
        }
        writeToCache(cachedFile, result);
        return result;
      }
    }
  }

  /** Instruments all files in the class file source. <BR>
   *  If a run of this method finishes successfully, all classes in the
   *  source were instrumented.
   *  @throws IOException if an exception occurs reading/writing cached results
   *  @throws InstrumentationException if the runtime jar file cannot be found
   */
  public void instrument() throws IOException, InstrumentationException, ClassNotFoundException {
    TraceInstrumenter traceInstrumenter;
    InstrumentationContext context;
    ClassFileSource source;
    IdentifierMapGenerator mapGenerator;
    IdentifierMap identityMap;
    
    //Instrument classes in place
    source = new DirectorySource(classDirectory, classDirectory);
    instrumentClassNames = getInstrumentClassNames(source);
    context = new InstrumentationContext(getCoreClassSet(),
                                         new HashSet(), instrumentClassNames, filter);
    mapGenerator = new IdentifierMapGenerator(context, source);
    identityMap = mapGenerator.generateIdentifierMap();
    context.setIdentifierMap(identityMap);
    traceInstrumenter = new TraceInstrumenter(getTraceExtensions(context), source, context);
    traceInstrumenter.instrument();
    source.saveModifiedVersion();
    identityMap.saveIdentifierMap(workingDirectory.getAbsolutePath() + File.separator + "identity.map");
  }

  public HashSet getInstrumentClassNames() {
    return instrumentClassNames;
  }
}
