/**
 * <p>Title: Bytecode instrumentation</p>
 * <p>Description: Tracing framework for java bytecode</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universit\uFFFDt Saarbr\uFFFDcken</p>
 * @author Valentin Dallmeier
 * @version 1.0
 */
package org.softevo.ample.sibrelib.extensions;

import org.apache.bcel.Constants;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.ClassGen;
import org.apache.bcel.generic.FieldGen;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionFactory;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.INVOKESPECIAL;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.Type;
import org.softevo.sibrelib.InstrumentationContext;
import org.softevo.sibrelib.extensions.TraceInstrumentationExtension;

/** Creates a tracer that instruments classes to trace
 *  information about the instance a method called
 *  was issued upon. <BR>
 */
public class InstanceTracer extends TraceInstrumentationExtension {

  /** The string that is traced when the method is static. <BR>
   */
  public static final String FIELDNAME_ID = "___id";

  /** Create a new instance. <BR>
   *  @param context the instrumentation context
   */
  public InstanceTracer(InstrumentationContext context) {
    super(context);
  }

  /** This method performs the neccessary changes to
   *  a java class that are needed in order to guarantee
   *  tracing of instances by their address. <BR>
   *  The following changes are made:
   *  <ol>
   *   <li>In case this class is a sub class of java.lang.Object,
   *       two fields are added:
   *       One holds a flag that signalizes the <code>hashCode</code> method
   *       to call its super class version (thus returning the object address)
   *       and the other one is used to store the retrieved value.
   *       Every constructor declared in the class is instrumented with code
   *       to set the flag, call the <code>hashCode</code> method,
   *       retrieve the address, store it in the <code>___add</code> field
   *       and reset the flag</li>
   *   <li>Every declared version of <code>hashCode</code> is instrumentated
   *       with code to call the super class version if the <code>___flag</code>
   *       field was set.</li>
   *  </ol>
   *  @param javaClass the class to process
   *  @param classGenerator the generator for the new class
   *  @param constantPoolGenerator the generator for the new constant pool
   *  @param factory the factory for instructions
   */
  public void postProcessJavaClass(JavaClass javaClass,
                                   ClassGen classGenerator,
                                   ConstantPoolGen constantPoolGenerator,
                                   InstructionFactory factory) {
    FieldGen fieldGen;
    Method[] methods;
    MethodGen methodGenerator;
    InstructionHandle[] handles;
    Instruction instruction;
    InstructionList instructions;
    int innerCounter = 0;
    INVOKESPECIAL invokeInstruction;
    boolean instrumentConstructor;

    methods = classGenerator.getMethods();
    //since no api class is instrumented we must assure
    //that we reach all classes whose super class is an
    //api class
    if (isTopInstrumentedClass(javaClass)) {
      //add the fields since this is the upper most class
      fieldGen = new FieldGen(org.apache.bcel.Constants.ACC_PUBLIC,
                              Type.INT, FIELDNAME_ID, constantPoolGenerator);
      classGenerator.addField(fieldGen.getField());
      methods = classGenerator.getMethods();
      //search all methods in order to find the constructor
      for (int counter = 0; counter < methods.length; counter++) {
        if (methods[counter].getName().equals("<init>")) {
          instrumentConstructor = true;
          methodGenerator = new MethodGen(methods[counter],
                                          javaClass.getClassName(),
                                          constantPoolGenerator);
          handles = methodGenerator.getInstructionList().getInstructionHandles();
          //we search all instructions to find the special invokations
          //these are the constructor calls
          //we need to take special precaution for constructor
          //calls that have new calls, e.g. this(new Test());
          //innerCounter = 0;
          for (innerCounter = 0; innerCounter < handles.length; innerCounter++) {
            instruction = handles[innerCounter].getInstruction();
            if (instruction instanceof INVOKESPECIAL) {
              invokeInstruction = (INVOKESPECIAL) instruction;
              //we found the super class constructor call
              if (invokeInstruction.getClassName(constantPoolGenerator).equals(javaClass.getSuperclassName())) {
                break;
              }
              //we found a call to a constructor in the same class
              //if this is the case we do not need to instrument the constructor
              if (invokeInstruction.getClassName(constantPoolGenerator).equals(javaClass.getClassName()) &&
                  invokeInstruction.getMethodName(constantPoolGenerator).equals("<init>")) {
                instrumentConstructor = false;
                break;
              }
            }
          }
          if (instrumentConstructor) {
            //find the next instruction after the invocation
            //(must exist in valid bytecode)
            innerCounter++;
            instructions = new InstructionList();
            instructions.append(InstructionFactory.createThis());
            instructions.append(factory.createInvoke("org.softevo.sibrelib.runtime.IDCreator",
                                                     "getNextId", Type.INT, Type.NO_ARGS,
                                                     Constants.INVOKESTATIC));
            instructions.append(factory.createFieldAccess(javaClass.
                                                          getClassName(),
                                                          FIELDNAME_ID,
                                                          Type.INT,
                                                          Constants.PUTFIELD));
            methodGenerator.getInstructionList().insert(handles[innerCounter],
                                                        instructions);
            methodGenerator.setMaxLocals();
            methodGenerator.setMaxStack();
            classGenerator.setMethodAt(methodGenerator.getMethod(), counter);
          }
        }
      }
    }
  }

  protected boolean isTopInstrumentedClass(JavaClass javaClass) {
    String superClassName;
    
    superClassName = javaClass.getSuperclassName();
    while (!context.isCoreClass(superClassName) && !superClassName.equals("java.lang.Object")) {
      if (context.isInstrumentedClass(superClassName)) {
        return false;
      } else {
        javaClass = Repository.lookupClass(superClassName);
        if (javaClass != null) {
          superClassName = javaClass.getSuperclassName();
        } else {
          return true;
        }
      }
    }
    return true;
  }
  
  /** This method gives extensions
   *  the possibility to forbid instrumentation
   *  of special methods that were added during the
   *  <code>processJavaClass</code> method.
   *  The default implementation simply returns <code>true</code>
   *  @param methodGenerator the generator for the method
   *  @return <code>true</code> if the method should be instrumentated
   *          with trace code,
   *          <code>false</code> otherwise
   */
  public boolean allowTraceInstrumentation(MethodGen methodGenerator) {
    return true;
  }

}
