/*
 *
 */
package org.softevo.ample.sibrelib.extensions;

import java.util.HashMap;

import org.apache.bcel.Constants;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ClassGen;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.InstructionFactory;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.PUSH;
import org.apache.bcel.generic.Type;
import org.softevo.sibrelib.InstrumentationContext;
import org.softevo.sibrelib.InstrumentationException;
import org.softevo.sibrelib.extensions.TraceInstrumentationExtension;

/**
 * @author dallmeier
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class SequenceTracer extends TraceInstrumentationExtension {

  /** The instance tracer. <BR>
   */
  protected InstanceTracer instanceTracer = null;
  
  /** The thread tracer. <BR>
   */
  protected ThreadTracer threadTracer = null;

  /** The name of the runtime class. <BR>
   */
  protected String runtimeClassName = null;
  
  public SequenceTracer(InstrumentationContext context) {
    super(context);
    instanceTracer = new InstanceTracer(context);
    threadTracer = new ThreadTracer(context);
  }

  /** Checks if neither thread nor instance tracer forbid
   *  instrumentation. <BR> 
   */
  public boolean allowTraceInstrumentation(MethodGen methodGenerator) {
    return (instanceTracer.allowTraceInstrumentation(methodGenerator) &&
        threadTracer.allowTraceInstrumentation(methodGenerator));
  }
  
  /** Processes a java class and adds special methods and
   *  fields needed for the instrumentation extension. <BR>
   *  The default implementation does not change anything.
   *  Since extensions may have processed the class before,
   *  all changes concerning methods and fields must be applied
   *  to the <code>classGenerator</code> variable passed (and <b>not</b>
   *  to the java class.
   *  This method is called <b>before</b> trace code is inserted
   *  into methods.
   *  @param javaClass the class to process
   *  @param classGenerator the generator for the new class
   *  @param constantPoolGenerator the generator for the new constant pool
   *  @param factory the factory for instructions
   */
  public void preProcessJavaClass(JavaClass javaClass,
                                  ClassGen classGenerator,
                                  ConstantPoolGen constantPoolGenerator,
                                  InstructionFactory factory) {
    threadTracer.preProcessJavaClass(javaClass, classGenerator, 
                                     constantPoolGenerator, factory);
  }
  
  
  
  public InstructionList createMethodStartCode(JavaClass javaClass,
                                               ConstantPoolGen constantPoolGenerator,
                                               MethodGen methodGenerator,
                                               InstructionFactory factory,
                                               HashMap identifierMap,
                                               String runtimeClassName) throws InstrumentationException {
    InstructionList result;
    Integer id;
    
    result = new InstructionList();
    result.append(factory.createInvoke(javaClass.getClassName(),
                                       ThreadTracer.METHODNAME_GETTHREADID,
                                       Type.INT, new Type[] {},
                                       Constants.INVOKESTATIC));
    if (methodGenerator.isStatic()) {
      result.append(new PUSH(constantPoolGenerator, -1));
    } else {
      result.append(factory.createLoad(Type.OBJECT, 0));
      result.append(factory.createFieldAccess(javaClass.getClassName(), InstanceTracer.FIELDNAME_ID,
                                              Type.INT, Constants.GETFIELD));
    }
    id = (Integer) identifierMap.get(javaClass.getClassName());
    if (id != null) {
      result.append(new PUSH(constantPoolGenerator, id.intValue()));
    } else {
      throw new InstrumentationException("Unable to lookup id for class " + javaClass.getClassName());
    }
    id = (Integer) identifierMap.get(javaClass.getClassName() + "#" +
                                     methodGenerator.getName() + "#" +
                                     methodGenerator.getSignature());
    if (id != null) {
      result.append(new PUSH(constantPoolGenerator, id.intValue()));
    } else {
      throw new InstrumentationException("Unable to lookup id for method " + 
                                         javaClass.getClassName() + "#" +
                                         methodGenerator.getName() + "#" +
                                         methodGenerator.getSignature());
    }
    result.append(factory.createInvoke(runtimeClassName,
                                       "writeEntry", Type.VOID,
                                       new Type[] {Type.INT, Type.INT, Type.INT},
                                       Constants.INVOKESTATIC));   
    return result;
  }

  protected MethodGen getFinalizer(JavaClass javaClass, 
                                   ClassGen classGenerator,
                                   ConstantPoolGen constantPoolGenerator,
                                   InstructionFactory factory,
                                   String runtimeClassName) {
    InstructionList result = new InstructionList();
    int id;
    
    MethodGen methodGenerator = new MethodGen(Constants.ACC_PUBLIC, Type.VOID, Type.NO_ARGS, 
                                              new String[] {  }, "finalize", 
                                              javaClass.getClassName(), result,
                                              constantPoolGenerator);
    result.append(factory.createInvoke(javaClass.getClassName(),
                                       ThreadTracer.METHODNAME_GETTHREADID,
                                       Type.INT, new Type[] {},
                                       Constants.INVOKESTATIC));
    if (methodGenerator.isStatic()) {
      result.append(new PUSH(constantPoolGenerator, -1));
    } else {
      result.append(factory.createLoad(Type.OBJECT, 0));
      result.append(factory.createFieldAccess(javaClass.getClassName(), InstanceTracer.FIELDNAME_ID,
                                              Type.INT, Constants.GETFIELD));
    }
    id = context.getIdentifierMap().getIdentifier(javaClass.getClassName());
    result.append(new PUSH(constantPoolGenerator, id));
    result.append(factory.createInvoke(runtimeClassName, 
                                       "removeInstance", 
                                       Type.VOID, 
                                       new Type[] { Type.INT, Type.INT, Type.INT }, 
                                       Constants.INVOKESTATIC));
    result.append(factory.createReturn(Type.VOID));
    methodGenerator.setMaxStack();
    methodGenerator.setMaxLocals();
    return methodGenerator;
  }

  protected void addFinalizer(ClassGen classGenerator, JavaClass javaClass, ConstantPoolGen constantPoolGenerator,
                              InstructionFactory factory, String runtimeClassName) {
    Method[] methods;
    boolean finalizerPresent = false;
    
    methods = classGenerator.getMethods();
    for(int counter = 0; counter < methods.length; counter++) {
      if (methods[counter].getName().equals("finalize") &&
          methods[counter].getReturnType().equals(Type.VOID) &&
          methods[counter].getArgumentTypes().length == 0 &&
          methods[counter].isProtected()) {
        finalizerPresent = true;
        break;
      }
    }
    if (!finalizerPresent) {
      classGenerator.addMethod(getFinalizer(javaClass, classGenerator, constantPoolGenerator, factory, runtimeClassName).getMethod());
    } else {
      System.out.println("Warning: Finalizer present in class " + javaClass.getClassName());
    }
    
  }
}
