/**
 * <p>Title: Bytecode instrumentation</p>
 * <p>Description: Tracing framework for java bytecode</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universitt Saarbrcken</p>
 * @author Valentin Dallmeier
 * @version 1.0
 */
package org.softevo.ample.sibrelib.extensions;

import org.apache.bcel.Constants;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.generic.BranchInstruction;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.ClassGen;
import org.apache.bcel.generic.InstructionFactory;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.PUSH;
import org.apache.bcel.generic.Type;
import org.apache.bcel.generic.ObjectType;
import org.softevo.sibrelib.InstrumentationContext;
import org.softevo.sibrelib.extensions.TraceInstrumentationExtension;

/** Traces an identifier for the active thread. <BR>
 */
public class ThreadTracer extends TraceInstrumentationExtension {

  /** Name for the <code>getThreadId</code> method. <BR>
   */
  public static final String METHODNAME_GETTHREADID = "___getThreadId";

  /** Create a new instance. <BR>
   *  @param context the instrumentation context
   */
  public ThreadTracer(InstrumentationContext context) {
    super(context);
  }

  /** This method simply adds a static method that
   *  determines an id for the active thread to
   *  every class whose super class is not instrumented. <BR>
   *  @param javaClass the class to process
   *  @param classGenerator the generator for the new class
   *  @param constantPoolGenerator the generator for the new constant pool
   *  @param factory the factory for instructions
   */
  public void preProcessJavaClass(JavaClass javaClass,
                               ClassGen classGenerator,
                               ConstantPoolGen constantPoolGenerator,
                               InstructionFactory factory) {
    if (!context.getInstrumentClassNames().contains(javaClass.getSuperclassName())) {
      classGenerator.addMethod(generateGetThreadIdMethod(javaClass, classGenerator,
                                                         constantPoolGenerator, factory).getMethod());
    }
  }

  /** Forbids instrumentation of the <code>getThreadId</code>
   *  method (because otherwise an endless loop occurs).<BR>
   *  @param methodGenerator the generator for the method
   *  @return <code>true</code> if the method should be instrumentated
   *          with trace code,
   *          <code>false</code> otherwise
   */
  public boolean allowTraceInstrumentation(MethodGen methodGenerator) {
    return !methodGenerator.getName().equals(METHODNAME_GETTHREADID);
  }

  /** Generates a static method that determines a thread id for
   *  the active thread. <BR>
   *  @param javaClass the class to process
   *  @param methodGenerator the generator for the method
   *  @param constantPoolGenerator the generator for the new constant pool
   *  @param factory the factory for instructions
   */
  private MethodGen generateGetThreadIdMethod(JavaClass javaClass,
                                              ClassGen classGenerator,
                                              ConstantPoolGen constantPoolGenerator,
                                              InstructionFactory factory) {
    InstructionList il = new InstructionList();
    MethodGen method = new MethodGen(Constants.ACC_PUBLIC | Constants.ACC_STATIC, 
                                     Type.INT, Type.NO_ARGS, 
                                     new String[] {  }, METHODNAME_GETTHREADID, 
                                     javaClass.getClassName(), il, constantPoolGenerator);
    InstructionHandle ih_0 = il.append(factory.createInvoke("java.lang.Thread", 
                                                            "currentThread", 
                                                            new ObjectType("java.lang.Thread"), 
                                                            Type.NO_ARGS, Constants.INVOKESTATIC));
    il.append(factory.createInvoke("java.lang.Object", "getClass", 
                                   new ObjectType("java.lang.Class"), 
                                   Type.NO_ARGS, Constants.INVOKEVIRTUAL));
    il.append(factory.createInvoke("java.lang.Class", "getName", 
                                   Type.STRING, Type.NO_ARGS, 
                                   Constants.INVOKEVIRTUAL));
    il.append(new PUSH(constantPoolGenerator, "java.lang.Thread"));
    il.append(factory.createInvoke("java.lang.String", "equals", 
                                   Type.BOOLEAN, new Type[] { Type.OBJECT }, 
                                   Constants.INVOKEVIRTUAL));
    BranchInstruction ifeq_14 = factory.createBranchInstruction(Constants.IFEQ, null);
    il.append(ifeq_14);
    InstructionHandle ih_17 = il.append(factory.createInvoke("java.lang.Thread", 
                                                             "currentThread", 
                                                             new ObjectType("java.lang.Thread"), 
                                                             Type.NO_ARGS, 
                                                             Constants.INVOKESTATIC));
    il.append(factory.createInvoke("java.lang.Object", "hashCode", Type.INT, 
                                   Type.NO_ARGS, Constants.INVOKEVIRTUAL));
    il.append(factory.createReturn(Type.INT));
    InstructionHandle ih_24 = il.append(factory.createInvoke("java.lang.Thread", 
                                                             "currentThread", 
                                                             new ObjectType("java.lang.Thread"), 
                                                             Type.NO_ARGS, Constants.INVOKESTATIC));
    //this unfortunately triggers an IncompatibleClassChangeError
    //so we rely on the same code as for Thread class
    //TODO Find a better solution
    /*
    il.append(factory.createFieldAccess(javaClass.getClassName(), InstanceTracer.FIELDNAME_ID, 
                                        Type.INT, Constants.GETSTATIC));
                                        */
    il.append(factory.createInvoke("java.lang.Object", "hashCode", Type.INT, 
                                   Type.NO_ARGS, Constants.INVOKEVIRTUAL));
    InstructionHandle ih_30 = il.append(factory.createReturn(Type.INT));
    ifeq_14.setTarget(ih_24);
    method.setMaxStack();
    method.setMaxLocals();
    return  method;
  }

}
